/* 
 *
 *  Copyright 1996 Claude Lecommandeur. See COPYRIGHT file for details
 *
 *  Author:  Claude Lecommandeur [ Claude.Lecommandeur@epfl.ch ][ August 1996 ]
 *
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>
#include <X.h>

#define WINDOWTYPE_NORMAL          0
#define WINDOWTYPE_DIALOG          1
#define WINDOWTYPE_POPUP           2
#define WINDOWTYPE_MODELESSDIALOG  3

#define DialogList_INCREMENT 10
static Window *DialogList;
static int     DialogListPtr;
static int     DialogListEnd;

static void removeFromDialogList(Window);
static void addToDialogList(Window);
static void setDeleteWindowProtocol(Window);

void kaffe_awt_simple_NativeDisplay_createWindow (NativeDisplay display,
						  NativeWindow window) {
    NativeWindow	parent;
    Window		win, pw;
    int			x, y, w, h;
    int			backrgb;
    int			type;
    int			color;
    unsigned long	valuemask;
    XSetWindowAttributes attributes;

    parent  = unhand(window)->parent;
    type    = unhand(window)->type;
    pw      = ((parent == (NativeWindow) 0) || type) ?
			RootWindow (dpy, 0) : unhand(parent)->id;
    x       = unhand(window)->x;
    y       = unhand(window)->y;
    w       = unhand(window)->width;
    h       = unhand(window)->height;
    backrgb = unhand(window)->background;
    color   = getcolor (backrgb);

/*
   If the window is a popup then the co-ordinates we have been passed are
   relative to the parent window. However we need them to be relative to
   the root window so we convert them here.
*/
    if (type == WINDOWTYPE_POPUP) {
	int    ret_x, ret_y;
	Window ret_w;
	
	LOCKDISPLAY ("XTranslateCoords");
	XTranslateCoordinates(dpy, unhand(parent)->id, RootWindow(dpy, 0),
			      x, y, &ret_x, &ret_y, &ret_w);
	RELEASEDISPLAY ("XTranslateCoords");
	x = ret_x;
	y = ret_y;
    }
    if (debug) fprintf (stderr,
		"createWindow : p = %x, x = %d, y = %d, w = %d, h = %d, back = %x",
		pw, x, y, w, h, backrgb);

    if (w == 0) w = 1; if (h == 0) h = 1;
    LOCKDISPLAY ("createWindow");
    valuemask = 0;
    attributes.background_pixel	= color;	valuemask |= CWBackPixel;
    attributes.border_pixel	= 0;		valuemask |= CWBorderPixel;
    
    win = XCreateWindow (dpy, pw, x, y, w, h, 0,
			CopyFromParent,			/* depth */
			(unsigned int) CopyFromParent,	/* class */
			(Visual *) CopyFromParent,	/* visual */
			valuemask, &attributes);

    if (!parent || type) {  /* frame or popup or dialog */
	XSizeHints	sizehints;
	XWMHints	wmhints;
	JavaString	jtitle = unhand(window)->title;
	char		*title;

	title = jtitle ? makeCString (jtitle) : (char*)strdup ("Untitled");
	XSetStandardProperties (dpy, win, title, title, None, NULL, 0, NULL);

	wmhints.flags		= InputHint | StateHint;
	wmhints.input		= True;
	wmhints.initial_state	= NormalState;
	XSetWMHints (dpy, win, &wmhints);
	setDeleteWindowProtocol(win);
    }

    /* Set the TransientFor hint on dialgs so that they willall be iconified
       with the parent */
    if (type == WINDOWTYPE_DIALOG || type == WINDOWTYPE_MODELESSDIALOG){
	LOCKDISPLAY ("selectEvents");
	XSetTransientForHint(dpy, win, unhand(parent)->id);
	RELEASEDISPLAY ("selectEvents");
	if (type == WINDOWTYPE_DIALOG) addToDialogList(win);
    }

    /* Popup windows have OverrideRedirect set so that they get ignored
       by thw window manager */
    if (type == WINDOWTYPE_POPUP) {
	XSetWindowAttributes xwa;
	
	xwa.override_redirect = True;
	XChangeWindowAttributes(dpy,win, CWOverrideRedirect, &xwa);
    }

    RELEASEDISPLAY ("createWindow");
    if (debug) fprintf (stderr, "createWindow : res = %x\n", win);
    unhand(window)->id = (long)win;
    if (debug) fprintf (stderr, ", res = %x\n", win);
}

void kaffe_awt_simple_NativeDisplay_destroyWindow (NativeDisplay display,
						   NativeWindow window) {
    Window win = (Window) unhand(window)->id;

    if (debug) fprintf (stderr, "destroyWindow : %x\n", win);

    if (unhand(window)->type == WINDOWTYPE_DIALOG)
	removeFromDialogList(win);

    LOCKDISPLAY ("destroyWindow");
    XDestroyWindow (dpy, win);
    RELEASEDISPLAY ("destroyWindow");

}

void kaffe_awt_simple_NativeDisplay_mapWindow (NativeDisplay display,
					       NativeWindow window) {
    Window win = (Window) unhand(window)->id;

    if (debug) fprintf (stderr, "mapWindow : %x\n", win);
    LOCKDISPLAY ("mapWindow");

    /* I really wanted to do this in the Java part of the Dialog stuff but
       there seems to be no way for a peer to communicate geometry changes
       to it's target class. :-(

       So, excuses over... If we are mapping a dialog window then this is
       where we centre it over its parent. It must be done here because the
       peer does not (necessarily) have any valid geometry information available
       to it at create time.
    */
      
    if (unhand(window)->type == WINDOWTYPE_DIALOG ||
	unhand(window)->type == WINDOWTYPE_MODELESSDIALOG) {

	Window            parent;
	Window            ret_w;
	XWindowAttributes our_xwa;
	XWindowAttributes parent_xwa;
	int x,y;

	XGetTransientForHint(dpy, win, &parent);

/* Get our geometry and our parent's. Convert the parent's co-ordinates to
   root-window co-ordinates so we can position the new dialog on the screen.
   Bear in mind that the parent window may not be a top-level shell.
 */
	XGetWindowAttributes(dpy, win,    &our_xwa);
	XGetWindowAttributes(dpy, parent, &parent_xwa);

	XTranslateCoordinates(dpy, parent, RootWindow(dpy, 0),
			      parent_xwa.x, parent_xwa.y,
			      &parent_xwa.x, &parent_xwa.y, &ret_w);
	
	x = parent_xwa.x + parent_xwa.width/2  - our_xwa.width/2;
	y = parent_xwa.y + parent_xwa.height/2 - our_xwa.height/2;

	XMoveWindow(dpy, win, x,y);
    }

    /* Actually map the window */
    XMapWindow (dpy, win);

    if (unhand(window)->type == WINDOWTYPE_POPUP)
    {
	Cursor backarrow = XCreateFontCursor(dpy, XC_right_ptr);
	
	XGrabPointer(dpy, win, False,
		     ButtonPressMask  | ButtonReleaseMask |
		     ButtonMotionMask | PointerMotionMask,
		     GrabModeAsync, GrabModeAsync,
		     None,
		     backarrow,
		     CurrentTime);
	XFreeCursor(dpy, backarrow);
    }
    RELEASEDISPLAY ("mapWindow");
}

void kaffe_awt_simple_NativeDisplay_unmapWindow (NativeDisplay display,
						 NativeWindow window) {
    Window win = (Window) unhand(window)->id;

    if (debug) fprintf (stderr, "unmapWindow : %x\n", win);
    LOCKDISPLAY ("unmapWindow");
    XUnmapWindow (dpy, win);
    RELEASEDISPLAY ("unmapWindow");
}

void kaffe_awt_simple_NativeDisplay_moveWindow (NativeDisplay display,
						NativeWindow window) {
    Window win = (Window) unhand(window)->id;
    int x, y;

    x = unhand(window)->x;
    y = unhand(window)->y;
    if (debug) fprintf (stderr, "moveWindow : %x, x = %d, y = %d\n", win, x, y);
    LOCKDISPLAY ("moveWindow");
    XMoveWindow (dpy, win, x, y);
    RELEASEDISPLAY ("moveWindow");
}

void kaffe_awt_simple_NativeDisplay_resizeWindow (NativeDisplay display,
						  NativeWindow window) {
    Window win = (Window) unhand(window)->id;
    int w, h;

    w = unhand(window)->width;
    h = unhand(window)->height;
    if (debug) fprintf (stderr, "resizeWindow : %x, w = %d, h = %d\n", win, w, h);
    if (w == 0) w = 1; if (h == 0) h = 1;
    LOCKDISPLAY ("resizeWindow");
    XResizeWindow (dpy, win, w, h);
    RELEASEDISPLAY ("resizeWindow");
}

void kaffe_awt_simple_NativeDisplay_reshapeWindow (NativeDisplay display,
						   NativeWindow window) {
    Window win = (Window) unhand(window)->id;
    int x, y, w, h;

    x = unhand(window)->x;
    y = unhand(window)->y;
    w = unhand(window)->width;
    h = unhand(window)->height;
    if (debug) fprintf (stderr, "reshapeWindow : %x, x = %d, y = %d, w = %d, h = %d\n",
			win, x, y, w, h);
    if (w == 0) w = 1; if (h == 0) h = 1;
    LOCKDISPLAY ("reshapeWindow");
    XMoveResizeWindow (dpy, win, x, y, w, h);
    RELEASEDISPLAY ("reshapeWindow");
}

void kaffe_awt_simple_NativeDisplay_raiseWindow (NativeDisplay display,
						 NativeWindow window) {
    Window win = (Window) unhand(window)->id;

    if (debug) fprintf (stderr, "raiseWindow : %x\n", win);
    LOCKDISPLAY ("raiseWindow");
    XRaiseWindow (dpy, win);
    RELEASEDISPLAY ("raiseWindow");
}

void kaffe_awt_simple_NativeDisplay_lowerWindow (NativeDisplay display,
						 NativeWindow window) {
    Window win = (Window) unhand(window)->id;

    if (debug) fprintf (stderr, "lowerWindow : %x\n", win);
    LOCKDISPLAY ("lowerWindow");
    XLowerWindow (dpy, win);
    RELEASEDISPLAY ("lowerWindow");
}

void kaffe_awt_simple_NativeDisplay_setBackground (NativeDisplay display,
						   NativeWindow window) {
    Window	win	= (Window) unhand(window)->id;
    int		rgb	= unhand(window)->background;
    int		color;

    if (debug) fprintf (stderr, "setBackground : win = %x, rgb = %x\n", win, rgb);
    color = getcolor (rgb);
    LOCKDISPLAY ("setBackground");
    XSetWindowBackground (dpy, win, color);
    RELEASEDISPLAY ("setBackground");
}

void kaffe_awt_simple_NativeDisplay_setForeground (NativeDisplay display,
						   NativeWindow window) {
    Window win	= (Window) unhand(window)->id;
    int rgb	= unhand(window)->background;
    XColor xcolor;
    Colormap	cmap;

    if (debug) fprintf (stderr, "setForeground : win = %x, rgb = %x\n", win, rgb);
}

void kaffe_awt_simple_NativeDisplay_setTitle (NativeDisplay display,
					      NativeWindow window) {
    Window win	= (Window) unhand(window)->id;
    char *title;
    
    title = makeCString (unhand(window)->title);
    if (debug) fprintf (stderr, "setTitle : win = %x\n", win);
    LOCKDISPLAY ("setTitle");
    XSetStandardProperties (dpy, win, title, title, None, NULL, 0, NULL);
    RELEASEDISPLAY ("setTitle");
}

void kaffe_awt_simple_NativeDisplay_selectEvents (NativeDisplay display,
						  NativeWindow window) {
    Window win = (Window) unhand(window)->id;
    int mask = unhand(window)->eventmask;

    if (debug) fprintf (stderr, "selectEvents : %x, mask = %x\n", win, mask);
    LOCKDISPLAY ("selectEvents");
    XSelectInput (dpy, win, mask);
    RELEASEDISPLAY ("selectEvents");
}

void kaffe_awt_simple_NativeDisplay_setInputFocus (NativeDisplay display,
						   NativeWindow window) {
    Window win = (Window) unhand(window)->id;

    if (debug) fprintf (stderr, "selectEvents : %x, mask = %x\n", win);
    LOCKDISPLAY ("selectEvents");
    XSetInputFocus (dpy, win, RevertToPointerRoot, CurrentTime);
    RELEASEDISPLAY ("selectEvents");
}

void kaffe_awt_simple_NativeDisplay_setResizeable (NativeDisplay display,
						   NativeWindow window,
                                                   jint resizeable) {
    XSizeHints        xsh;
    Window            win = unhand(window)->id;
    
    if (debug) fprintf (stderr, "setResizeable : %x: = %d\n",
			unhand(window)->id, resizeable);

    if (resizeable) {
	XWindowAttributes xwa;
	
	XGetWindowAttributes(dpy, (Window)RootWindow(dpy, 0), &xwa);
	xsh.min_width = 0;
	xsh.max_width = xwa.width;
	xsh.min_height = 0;
	xsh.max_height = xwa.height;
    }
    else {
	int width  = unhand(window)->width;
	int height = unhand(window)->height;

	xsh.min_width  = width;
	xsh.max_width  = width;
	xsh.min_height = height;
	xsh.max_height = height;
    }
    if (debug) fprintf(stderr, "size hints: %d,%d, %d,%d\n",
		       xsh.min_width, xsh.max_width,
				 xsh.min_height, xsh.max_height);
    xsh.flags = PMinSize | PMaxSize;
    XSetWMNormalHints(dpy, win, &xsh);
}



static void addToDialogList(Window win) {
    if (DialogList == NULL) /* Create for the first time */
    {
	DialogList = (Window*)malloc(sizeof(Window)+DialogList_INCREMENT);
	DialogListPtr = 0;
	DialogListEnd = DialogList_INCREMENT;
    }
    if (DialogListPtr == DialogListEnd) /* Grow the list if needed */
    {
	DialogListEnd += DialogList_INCREMENT;
	DialogList = (Window*)realloc(DialogList, sizeof(Window)*DialogListEnd);
    }
    DialogList[DialogListPtr++] = win;
}

static void removeFromDialogList(Window win) {
    int i;

    /* When a window is removed from the dialog list, it implicitly
       removes all windows higher up the list.
       This should take care of any left-over windows in the list
       Not that there should be any of course !
     */
    for (i=DialogListPtr-1; i>=0 && DialogList[i] != win; i--)
	;

    if (i < DialogListPtr)
    {
	DialogListPtr = i;
    }
}

/* Returns True if the window is a control on the top dialog or actually
   is the top dialog, False if not.
 */
int isOnTopDialog(Window win) {
    int i;
    Window      parent;
    Window      *children;
    Window      root = 0;
    Status      st;
    int         numChildren;
    XWindowAttributes xwa;

    if (DialogListPtr == 0) return True; /* No dialogs present */

/* Follow the window tree upwards looking for the top dialog window.
   If we reach the RootWindow then the control must be on some other
   dialog/window so we can return False.
 */   
    while (win != root && win != DialogList[DialogListPtr-1] && st) {
	st = XQueryTree(dpy, win,
			&root, &parent,
			&children, &numChildren);
	if (st) {
	    if (children != NULL) XFree(children);
	    XGetWindowAttributes(dpy, win, &xwa);
	    win = parent;
	}
    }

    /* We need to check for override_redirect windows too so that popups of
       Choice are handled correctly.
       The simple check will not be a problem because a choice popup cannot
       be created unless the main choice widget active
     */
    if ((win == root && !xwa.override_redirect) || !st)
	return False;
    else
	return True;
}

static void setDeleteWindowProtocol(Window win) {
    Atom *old_protocols = NULL;
    int  count;

    if (XGetWMProtocols(dpy, win, &old_protocols, &count)) {
	Atom *new_protocols;
	
	new_protocols = (Atom*)malloc((count+1)*sizeof(Atom));
	memcpy(new_protocols, old_protocols, count*sizeof(Atom));
	new_protocols[count] = XInternAtom(dpy, "WM_DELETE_WINDOW", False);
	
	XSetWMProtocols(dpy, win, new_protocols, count+1);
	XFree(old_protocols);
	free(new_protocols);
    }
    else {
	Atom new_protocols[1];
	
	new_protocols[0] = XInternAtom(dpy, "WM_DELETE_WINDOW", False);
	XSetWMProtocols(dpy, win, new_protocols, 1);
    }
}
